/**
 * @copyright Cube Dev, Inc.
 * @license Apache-2.0
 * @fileoverview The `ClickHouseDriver` and related types declaration.
 */
import { BaseDriver, DownloadQueryResultsOptions, DownloadQueryResultsResult, DownloadTableCSVData, DriverCapabilities, DriverInterface, QueryOptions, QuerySchemasResult, StreamOptions, StreamTableDataWithTypes, TableColumn, TableQueryResult, TableStructure, UnloadOptions } from '@cubejs-backend/base-driver';
import { ClickHouseClient } from '@clickhouse/client';
import type { ClickHouseSettings, ResponseJSON } from '@clickhouse/client';
export interface ClickHouseDriverOptions {
    host?: string;
    port?: string;
    username?: string;
    password?: string;
    protocol?: string;
    database?: string;
    readOnly?: boolean;
    /**
     * Timeout in milliseconds for requests to ClickHouse.
     * Default is 10 minutes
     */
    requestTimeout?: number;
    /**
     * Data source name.
     */
    dataSource?: string;
    /**
     * Max pool size value for the [cube]<-->[db] pool.
     */
    maxPoolSize?: number;
    /**
     * Time to wait for a response from a connection after validation
     * request before determining it as not valid. Default - 10000 ms.
     */
    testConnectionTimeout?: number;
}
interface ClickhouseDriverExportRequiredAWS {
    bucketType: 's3';
    bucketName: string;
    region: string;
}
interface ClickhouseDriverExportKeySecretAWS extends ClickhouseDriverExportRequiredAWS {
    keyId: string;
    secretKey: string;
}
interface ClickhouseDriverExportAWS extends ClickhouseDriverExportKeySecretAWS {
}
type ClickHouseDriverConfig = {
    url: string;
    username: string;
    password: string;
    readOnly: boolean;
    database: string;
    requestTimeout: number;
    exportBucket: ClickhouseDriverExportAWS | null;
    compression: {
        response?: boolean;
        request?: boolean;
    };
    clickhouseSettings: ClickHouseSettings;
};
export declare class ClickHouseDriver extends BaseDriver implements DriverInterface {
    /**
     * Returns default concurrency value.
     */
    static getDefaultConcurrency(): number;
    protected readonly client: ClickHouseClient;
    protected readonly readOnlyMode: boolean;
    protected readonly config: ClickHouseDriverConfig;
    /**
     * Class constructor.
     */
    constructor(config?: ClickHouseDriverOptions);
    protected withCancel<T>(fn: (con: ClickHouseClient, queryId: string, signal: AbortSignal) => Promise<T>): Promise<T>;
    protected createClient(maxPoolSize: number): ClickHouseClient;
    testConnection(): Promise<void>;
    readOnly(): boolean;
    query<R = unknown>(query: string, values: unknown[]): Promise<R[]>;
    protected queryResponse(query: string, values: unknown[]): Promise<ResponseJSON<Record<string, unknown>>>;
    protected normaliseResponse<R = unknown>(res: ResponseJSON<Record<string, unknown>>): Array<R>;
    release(): Promise<void>;
    informationSchemaQuery(): string;
    protected getTablesForSpecificSchemasQuery(schemasPlaceholders: string): string;
    protected getColumnsForSpecificTablesQuery(conditionString: string): string;
    protected getColumnNameForSchemaName(): string;
    protected getColumnNameForTableName(): string;
    getSchemas(): Promise<QuerySchemasResult[]>;
    stream(query: string, values: unknown[], { highWaterMark }: StreamOptions): Promise<StreamTableDataWithTypes>;
    downloadQueryResults(query: string, values: unknown[], options: DownloadQueryResultsOptions): Promise<DownloadQueryResultsResult>;
    toGenericType(columnType: string): string;
    createSchemaIfNotExists(schemaName: string): Promise<void>;
    getTablesQuery(schemaName: string): Promise<TableQueryResult[]>;
    dropTable(tableName: string, _options?: QueryOptions): Promise<void>;
    protected getExportBucket(dataSource: string): ClickhouseDriverExportAWS | null;
    isUnloadSupported(): Promise<boolean>;
    /**
     * Returns an array of queried fields meta info.
     */
    queryColumnTypes(sql: string, params: unknown[]): Promise<TableStructure>;
    createTableRaw(query: string): Promise<void>;
    createTable(quotedTableName: string, columns: TableColumn[]): Promise<void>;
    /**
     * We use unloadWithoutTempTable strategy
     */
    unload(_tableName: string, options: UnloadOptions): Promise<DownloadTableCSVData>;
    unloadFromQuery(sql: string, params: unknown[], _options: UnloadOptions): Promise<DownloadTableCSVData>;
    capabilities(): DriverCapabilities;
    command(query: string): Promise<void>;
    insert(table: string, values: Array<Array<unknown>>): Promise<void>;
}
export {};
//# sourceMappingURL=ClickHouseDriver.d.ts.map