"use strict";
/**
 * @copyright Cube Dev, Inc.
 * @license Apache-2.0
 * @fileoverview The `ClickHouseDriver` and related types declaration.
 */
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.ClickHouseDriver = void 0;
const shared_1 = require("@cubejs-backend/shared");
const base_driver_1 = require("@cubejs-backend/base-driver");
const node_stream_1 = require("node:stream");
const client_1 = require("@clickhouse/client");
const uuid_1 = require("uuid");
const sqlstring_1 = __importDefault(require("sqlstring"));
const HydrationStream_1 = require("./HydrationStream");
const SUPPORTED_BUCKET_TYPES = ['s3'];
const ClickhouseTypeToGeneric = {
    enum: 'text',
    string: 'text',
    datetime: 'timestamp',
    datetime64: 'timestamp',
    date: 'date',
    decimal: 'decimal',
    // integers
    int8: 'int',
    int16: 'int',
    int32: 'int',
    int64: 'bigint',
    // unsigned int
    uint8: 'int',
    uint16: 'int',
    uint32: 'int',
    uint64: 'bigint',
    // floats
    float32: 'float',
    float64: 'double',
    // We don't support enums
    enum8: 'text',
    enum16: 'text',
};
class ClickHouseDriver extends base_driver_1.BaseDriver {
    /**
     * Returns default concurrency value.
     */
    static getDefaultConcurrency() {
        return 10;
    }
    // ClickHouseClient has internal pool of several sockets, no need for generic-pool
    client;
    readOnlyMode;
    config;
    /**
     * Class constructor.
     */
    constructor(config = {}) {
        super({
            testConnectionTimeout: config.testConnectionTimeout,
        });
        const dataSource = config.dataSource ?? (0, shared_1.assertDataSource)('default');
        const host = config.host ?? (0, shared_1.getEnv)('dbHost', { dataSource });
        const port = config.port ?? (0, shared_1.getEnv)('dbPort', { dataSource }) ?? 8123;
        const protocol = config.protocol ?? ((0, shared_1.getEnv)('dbSsl', { dataSource }) ? 'https:' : 'http:');
        const url = `${protocol}//${host}:${port}`;
        const username = config.username ?? (0, shared_1.getEnv)('dbUser', { dataSource });
        const password = config.password ?? (0, shared_1.getEnv)('dbPass', { dataSource });
        const database = config.database ?? (0, shared_1.getEnv)('dbName', { dataSource }) ?? 'default';
        // TODO this is a bit inconsistent with readOnly
        this.readOnlyMode = (0, shared_1.getEnv)('clickhouseReadOnly', { dataSource });
        // Expect that getEnv('dbQueryTimeout') will always return a value
        const requestTimeoutEnv = (0, shared_1.getEnv)('dbQueryTimeout', { dataSource }) * 1000;
        const requestTimeout = config.requestTimeout ?? requestTimeoutEnv;
        this.config = {
            url,
            username,
            password,
            database,
            exportBucket: this.getExportBucket(dataSource),
            readOnly: !!config.readOnly,
            requestTimeout,
            compression: {
                // Response compression can't be enabled for a user with readonly=1, as ClickHouse will not allow settings modifications for such user.
                response: this.readOnlyMode ? false : (0, shared_1.getEnv)('clickhouseCompression', { dataSource }),
                request: (0, shared_1.getEnv)('clickhouseCompression', { dataSource }),
            },
            clickhouseSettings: {
                // If ClickHouse user's permissions are restricted with "readonly = 1",
                // change settings queries are not allowed. Thus, "join_use_nulls" setting
                // can not be changed
                ...(this.readOnlyMode ? {} : { join_use_nulls: 1 }),
            },
        };
        const maxPoolSize = config.maxPoolSize ?? (0, shared_1.getEnv)('dbMaxPoolSize', { dataSource }) ?? 8;
        this.client = this.createClient(maxPoolSize);
    }
    withCancel(fn) {
        const queryId = (0, uuid_1.v4)();
        const abortController = new AbortController();
        const { signal } = abortController;
        const promise = (async () => {
            const pingResult = await this.client.ping();
            if (!pingResult.success) {
                // TODO replace string formatting with proper cause
                // pingResult.error can be AggregateError when ClickHouse hostname resolves to multiple addresses
                let errorMessage = pingResult.error.toString();
                if (pingResult.error instanceof AggregateError) {
                    errorMessage = `Aggregate error: ${pingResult.error.message}; errors: ${pingResult.error.errors.join('; ')}`;
                }
                throw new Error(`Connection check failed: ${errorMessage}`);
            }
            signal.throwIfAborted();
            // Queries sent by `fn` can hit a timeout error, would _not_ get killed, and continue running in ClickHouse
            // TODO should we kill those as well?
            const result = await fn(this.client, queryId, signal);
            signal.throwIfAborted();
            return result;
        })();
        promise.cancel = async () => {
            abortController.abort();
            // Use separate client for kill query, usual pool may be busy
            const killClient = this.createClient(1);
            try {
                await killClient.command({
                    query: `KILL QUERY WHERE query_id = '${queryId}'`,
                });
            }
            finally {
                await killClient.close();
            }
        };
        return promise;
    }
    createClient(maxPoolSize) {
        return (0, client_1.createClient)({
            url: this.config.url,
            username: this.config.username,
            password: this.config.password,
            database: this.config.database,
            compression: this.config.compression,
            clickhouse_settings: this.config.clickhouseSettings,
            request_timeout: this.config.requestTimeout,
            max_open_connections: maxPoolSize,
        });
    }
    async testConnection() {
        await this.query('SELECT 1', []);
    }
    readOnly() {
        return (this.config.readOnly != null || this.readOnlyMode) ?
            (!!this.config.readOnly || this.readOnlyMode) :
            true;
    }
    async query(query, values) {
        const response = await this.queryResponse(query, values);
        return this.normaliseResponse(response);
    }
    queryResponse(query, values) {
        const formattedQuery = sqlstring_1.default.format(query, values);
        return this.withCancel(async (connection, queryId, signal) => {
            try {
                const format = 'JSON';
                const resultSet = await connection.query({
                    query: formattedQuery,
                    query_id: queryId,
                    format,
                    clickhouse_settings: this.config.clickhouseSettings,
                    abort_signal: signal,
                });
                // response_headers['x-clickhouse-format'] is optional, but if it exists,
                // it should match the requested format.
                if (resultSet.response_headers['x-clickhouse-format'] && resultSet.response_headers['x-clickhouse-format'] !== format) {
                    throw new Error(`Unexpected x-clickhouse-format in response: expected ${format}, received ${resultSet.response_headers['x-clickhouse-format']}`);
                }
                // We used format JSON, so we expect each row to be Record with column names as keys
                const results = await resultSet.json();
                return results;
            }
            catch (e) {
                // TODO replace string formatting with proper cause
                throw new Error(`Query failed: ${e}; query id: ${queryId}`);
            }
        });
    }
    normaliseResponse(res) {
        if (res.data) {
            const meta = (res.meta ?? []).reduce((state, element) => ({ [element.name]: element, ...state }), {});
            // TODO maybe use row-based format here as well?
            res.data.forEach((row) => {
                (0, HydrationStream_1.transformRow)(row, meta);
            });
        }
        return res.data;
    }
    async release() {
        await this.client.close();
    }
    informationSchemaQuery() {
        return `
      SELECT name as column_name,
             table as table_name,
             database as table_schema,
             type as data_type
        FROM system.columns
       WHERE database = '${this.config.database}'
    `;
    }
    getTablesForSpecificSchemasQuery(schemasPlaceholders) {
        const query = `
      SELECT database as schema_name,
            name as table_name
      FROM system.tables
      WHERE database IN (${schemasPlaceholders})
    `;
        return query;
    }
    getColumnsForSpecificTablesQuery(conditionString) {
        const query = `
      SELECT name as ${this.quoteIdentifier('column_name')},
             table as ${this.quoteIdentifier('table_name')},
             database as ${this.quoteIdentifier('schema_name')},
             type as ${this.quoteIdentifier('data_type')}
      FROM system.columns
      WHERE ${conditionString}
    `;
        return query;
    }
    getColumnNameForSchemaName() {
        return 'database';
    }
    getColumnNameForTableName() {
        return 'table';
    }
    async getSchemas() {
        return [{ schema_name: this.config.database }];
    }
    async stream(query, values, 
    // eslint-disable-next-line @typescript-eslint/no-unused-vars
    { highWaterMark }) {
        // Use separate client for this long-living query
        const client = this.createClient(1);
        const queryId = (0, uuid_1.v4)();
        try {
            const formattedQuery = sqlstring_1.default.format(query, values);
            const format = 'JSONCompactEachRowWithNamesAndTypes';
            const resultSet = await client.query({
                query: formattedQuery,
                query_id: queryId,
                format,
                clickhouse_settings: this.config.clickhouseSettings,
            });
            if (resultSet.response_headers['x-clickhouse-format'] !== format) {
                throw new Error(`Unexpected x-clickhouse-format in response: expected ${format}, received ${resultSet.response_headers['x-clickhouse-format']}`);
            }
            // Array<unknown> is okay, because we use fixed JSONCompactEachRowWithNamesAndTypes format
            // And each row after first two will look like this: [42, "hello", [0,1]]
            // https://clickhouse.com/docs/en/interfaces/formats#jsoncompacteachrowwithnamesandtypes
            const resultSetStream = resultSet.stream();
            const allRowsIter = (async function* allRowsIter() {
                for await (const rowsBatch of resultSetStream) {
                    for (const row of rowsBatch) {
                        yield row.json();
                    }
                }
            }());
            const first = await allRowsIter.next();
            if (first.done) {
                throw new Error('Unexpected stream end before row with names');
            }
            // JSONCompactEachRowWithNamesAndTypes: expect first row to be column names as string
            const names = first.value;
            const second = await allRowsIter.next();
            if (second.done) {
                throw new Error('Unexpected stream end before row with types');
            }
            // JSONCompactEachRowWithNamesAndTypes: expect first row to be column names as string
            const types = second.value;
            if (names.length !== types.length) {
                throw new Error(`Unexpected names and types length mismatch; names ${names.length} vs types ${types.length}`);
            }
            const dataRowsIter = (async function* () {
                for await (const row of allRowsIter) {
                    yield (0, HydrationStream_1.transformStreamRow)(row, names, types);
                }
            }());
            const rowStream = node_stream_1.Readable.from(dataRowsIter);
            return {
                rowStream,
                types: names.map((name, idx) => {
                    const type = types[idx];
                    return {
                        name,
                        type: this.toGenericType(type),
                    };
                }),
                release: async () => {
                    await client.close();
                }
            };
        }
        catch (e) {
            await client.close();
            // TODO replace string formatting with proper cause
            throw new Error(`Stream query failed: ${e}; query id: ${queryId}`);
        }
    }
    async downloadQueryResults(query, values, options) {
        if ((options ?? {}).streamImport) {
            return this.stream(query, values, options);
        }
        const response = await this.queryResponse(query, values);
        return {
            rows: this.normaliseResponse(response),
            types: (response.meta ?? []).map((field) => ({
                name: field.name,
                type: this.toGenericType(field.type),
            })),
        };
    }
    toGenericType(columnType) {
        if (columnType.toLowerCase() in ClickhouseTypeToGeneric) {
            return ClickhouseTypeToGeneric[columnType.toLowerCase()];
        }
        /**
         * Example of types:
         *
         * Int64
         * Nullable(Int64) / Nullable(String)
         * Nullable(DateTime('UTC'))
         */
        if (columnType.includes('(')) {
            const types = columnType.toLowerCase().match(/([a-z0-9']+)/g);
            if (types) {
                for (const type of types) {
                    if (type in ClickhouseTypeToGeneric) {
                        return ClickhouseTypeToGeneric[type];
                    }
                }
            }
        }
        return super.toGenericType(columnType);
    }
    async createSchemaIfNotExists(schemaName) {
        await this.command(`CREATE DATABASE IF NOT EXISTS ${schemaName}`);
    }
    getTablesQuery(schemaName) {
        return this.query('SELECT name as table_name FROM system.tables WHERE database = ?', [schemaName]);
    }
    async dropTable(tableName, _options) {
        await this.command(`DROP TABLE ${tableName}`);
    }
    getExportBucket(dataSource) {
        const requiredExportBucket = {
            bucketType: (0, shared_1.getEnv)('dbExportBucketType', {
                supported: SUPPORTED_BUCKET_TYPES,
                dataSource,
            }),
            bucketName: (0, shared_1.getEnv)('dbExportBucket', { dataSource }),
            region: (0, shared_1.getEnv)('dbExportBucketAwsRegion', { dataSource }),
        };
        const exportBucket = {
            ...requiredExportBucket,
            keyId: (0, shared_1.getEnv)('dbExportBucketAwsKey', { dataSource }),
            secretKey: (0, shared_1.getEnv)('dbExportBucketAwsSecret', { dataSource }),
        };
        if (exportBucket.bucketType) {
            if (!SUPPORTED_BUCKET_TYPES.includes(exportBucket.bucketType)) {
                throw new Error(`Unsupported EXPORT_BUCKET_TYPE, supported: ${SUPPORTED_BUCKET_TYPES.join(',')}`);
            }
            // Make sure the required keys are set
            const emptyRequiredKeys = Object.keys(requiredExportBucket)
                .filter((key) => requiredExportBucket[key] === undefined);
            if (emptyRequiredKeys.length) {
                throw new Error(`Unsupported configuration exportBucket, some configuration keys are empty: ${emptyRequiredKeys.join(',')}`);
            }
            return exportBucket;
        }
        return null;
    }
    async isUnloadSupported() {
        return !!this.config.exportBucket;
    }
    /**
     * Returns an array of queried fields meta info.
     */
    async queryColumnTypes(sql, params) {
        const columns = await this.query(`DESCRIBE ${sql}`, params);
        if (!columns) {
            throw new Error('Unable to describe table');
        }
        return columns.map((column) => ({
            name: column.name,
            type: this.toGenericType(column.type),
        }));
    }
    // This is only for use in tests
    async createTableRaw(query) {
        await this.command(query);
    }
    async createTable(quotedTableName, columns) {
        const createTableSql = this.createTableSql(quotedTableName, columns);
        try {
            await this.command(createTableSql);
        }
        catch (e) {
            // TODO replace string formatting with proper cause
            throw new Error(`Create table failed: ${e}`);
        }
    }
    /**
     * We use unloadWithoutTempTable strategy
     */
    async unload(_tableName, options) {
        if (!options.query?.sql) {
            throw new Error('Query must be defined in options');
        }
        return this.unloadFromQuery(options.query?.sql, options.query?.params, options);
    }
    async unloadFromQuery(sql, params, _options) {
        if (!this.config.exportBucket) {
            throw new Error('Unload is not configured');
        }
        const types = await this.queryColumnTypes(`(${sql})`, params);
        const { bucketName, path } = this.parseBucketUrl(this.config.exportBucket.bucketName);
        const exportPrefix = path ? `${path}/${(0, uuid_1.v4)()}` : (0, uuid_1.v4)();
        const formattedQuery = sqlstring_1.default.format(`
      INSERT INTO FUNCTION
         s3(
             'https://${bucketName}.s3.${this.config.exportBucket.region}.amazonaws.com/${exportPrefix}/export.csv.gz',
             '${this.config.exportBucket.keyId}',
             '${this.config.exportBucket.secretKey}',
             'CSV'
          )
      ${sql}
    `, params);
        await this.command(formattedQuery);
        const csvFile = await this.extractUnloadedFilesFromS3({
            credentials: {
                accessKeyId: this.config.exportBucket.keyId,
                secretAccessKey: this.config.exportBucket.secretKey,
            },
            region: this.config.exportBucket.region,
        }, bucketName, exportPrefix);
        return {
            csvFile,
            types,
            csvNoHeader: true,
            // Can be controlled via SET format_csv_delimiter
            csvDelimiter: ','
        };
    }
    capabilities() {
        return {
            unloadWithoutTempTable: true,
            incrementalSchemaLoading: true,
        };
    }
    // This is not part of a driver interface, and marked public only for testing
    async command(query) {
        await this.withCancel(async (connection, queryId, signal) => {
            await connection.command({
                query,
                query_id: queryId,
                abort_signal: signal,
            });
        });
    }
    // This is not part of a driver interface, and marked public only for testing
    async insert(table, values) {
        await this.withCancel(async (connection, queryId, signal) => {
            await connection.insert({
                table,
                values,
                format: 'JSONCompactEachRow',
                query_id: queryId,
                abort_signal: signal,
            });
        });
    }
}
exports.ClickHouseDriver = ClickHouseDriver;
//# sourceMappingURL=ClickHouseDriver.js.map